/*
   Copyright (c) CNRS,
   @author : Alireza NAKHAEI   anakhaei@laas.fr

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
modification,
   are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
notice,
   this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
   * Neither the name of the <ORGANIZATION> nor the names of its
contributors
   may be used to endorse or promote products derived from this software
without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS
   OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY
   AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER
   OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
   IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

/*****************************************
 INCLUDES
*******************************************/

#include <iostream>
#include "hppDynamicObstacle/hppDynamicObstacleKCD.h"
#include "hppDynamicObstacle/hppDynamicEnvHandler.h"

// Select verbosity at configuration by setting CXXFLAGS="... -DDEBUG=[1 or 2]"
#if DEBUG==2
#define ODEBUG2(x) std::cout << "hppDynamicObstacleKCD:" << x << std::endl
#define ODEBUG1(x) std::cerr << "hppDynamicObstacleKCD:" << x << std::endl
#elif DEBUG==1
#define ODEBUG2(x)
#define ODEBUG1(x) std::cerr << "hppDynamicObstacleKCD:" << x << std::endl
#else
#define ODEBUG2(x)
#define ODEBUG1(x)
#endif


unsigned int ChppDynamicObstacleKCD::nbObject = 0;

// ==========================================================================
ChppDynamicObstacleKCD::ChppDynamicObstacleKCD ( CkcdObjectShPtr inKCDObject,ChppDynamicEnvHandlerShPtr inDynamicEnvHandler ) :ChppDynamicObstacle ( inDynamicEnvHandler )
{
	attKCDObject=inKCDObject;
	nbObject++;
	ODEBUG2 ( "Constructor of ChppDynamicObstacleKCD. number of objects"<< nbObject );
}
// ==========================================================================
ChppDynamicObstacleKCD::~ChppDynamicObstacleKCD()
{
	nbObject--;
	ODEBUG2 ( "---------------------------------------------------------------------------------------------------" );
	ODEBUG2 ( "      Destructor of class ChppDynamicObstacleKCD. number of objects:" << nbObject );
	ODEBUG2 ( "--------------------------------------------------------------------------------------------------" );
}
// ==========================================================================
ChppDynamicObstacleKCDShPtr ChppDynamicObstacleKCD::create ( CkcdObjectShPtr inKCDObject,ChppDynamicEnvHandlerShPtr inDynamicEnvHandler )
{
	ChppDynamicObstacleKCD*  flatPtr = new ChppDynamicObstacleKCD ( inKCDObject, inDynamicEnvHandler );
	ChppDynamicObstacleKCDShPtr flatShPtr ( flatPtr );
	ChppDynamicObstacleKCDWkPtr flatWkPtr ( flatShPtr );

	if ( flatPtr->init ( flatWkPtr ) != KD_OK ) flatShPtr.reset();

	return flatShPtr;
}
// ==========================================================================
ktStatus ChppDynamicObstacleKCD::init ( const ChppDynamicObstacleKCDWkPtr &inWeakPtr )
{
	ktStatus success = this->ChppDynamicObstacle::init ( inWeakPtr ) ;
	attWeakPtr = inWeakPtr;

	return success;
}
// ==========================================================================
ktStatus ChppDynamicObstacleKCD::computeModifiedWorkSpaceCells()
{
	ODEBUG2 ( "compute Modified WorkSpace Cells .... " );
	double xMin=999999;
	double yMin=999999;
	double zMin=999999;

	double xMax=-999999;
	double yMax=-999999;
	double zMax=-999999;
	axisAlignedBoundingBox ( xMin, yMin, zMin, xMax, yMax, zMax );

	double z=zMin;
	bool flagZ=true;
	while ( z<=zMax )
	{
		double y=yMin;
		bool flagY=true;
		while ( y<=yMax )
		{
			double x=xMin;
			bool flagX=true;
			while ( x<=xMax )
			{
				getDynamicEnvHandler().lock()->posToCell ( x,y,z )->updateFlag ( true );
				x=x+getDynamicEnvHandler().lock()->getCellSize();
				if ( x>xMax && flagX==true )
				{
					x=xMax;
					flagX=false;
				}
			}
			y=y+getDynamicEnvHandler().lock()->getCellSize() ;
			if ( y>yMax && flagY==true  )
			{
				y=yMax;
				flagY=false;
			}
		}
		z=z+getDynamicEnvHandler().lock()->getCellSize() ;
		if ( z>zMax  && flagZ==true )
		{
			z=zMax;
			flagZ=false;
		}
	}

	ODEBUG2 ( ".....compute Modified WorkSpace Cells was finished. " );
	return KD_OK;
}
// ==========================================================================
CkcdObjectShPtr ChppDynamicObstacleKCD::getModifiedObstacle()
{
	return attKCDObject.lock();
}
// ==========================================================================
ktStatus ChppDynamicObstacleKCD::axisAlignedBoundingBox ( double& xMin, double& yMin, double& zMin, double& xMax, double& yMax, double& zMax ) const
{
	ODEBUG2 ( "axisAlignedBoundingBox .... " );

	xMin=999999;
	yMin=999999;
	zMin=999999;

	xMax=-999999;
	yMax=-999999;
	zMax=-999999;

	double x,y,z;

	attKCDObject.lock()->boundingBox()->getHalfLengths ( x, y, z ) ;

	/*Matrices absolute et relative*/
	CkitMat4 matrixAbsolutePosition;
	CkitMat4 matrixRelativePosition;
	attKCDObject.lock()->getAbsolutePosition ( matrixAbsolutePosition );
	attKCDObject.lock()->boundingBox()->getRelativePosition ( matrixRelativePosition );

	/*Creer les points et change position points*/
	CkitMat4 matrixChangePosition = matrixAbsolutePosition*matrixRelativePosition;

	CkitPoint3 position[8];

	position[0]=matrixChangePosition*CkitPoint3 ( x, y, z );
	position[1]=matrixChangePosition*CkitPoint3 ( x, y,-z );
	position[2]=matrixChangePosition*CkitPoint3 ( x,-y, z );
	position[3]=matrixChangePosition*CkitPoint3 ( -x, y, z );
	position[4]=matrixChangePosition*CkitPoint3 ( x,-y,-z );
	position[5]=matrixChangePosition*CkitPoint3 ( -x,-y, z );
	position[6]=matrixChangePosition*CkitPoint3 ( -x, y,-z );
	position[7]=matrixChangePosition*CkitPoint3 ( -x,-y,-z );

	for ( int i=0; i<8; i++ )
	{
		if ( ( position[i] ) [0]<xMin )
		{
			xMin= ( position[i] ) [0];
		}
		if ( ( position[i] ) [1]<yMin )
		{
			yMin= ( position[i] ) [1];
		}
		if ( ( position[i] ) [2]<zMin )
		{
			zMin= ( position[i] ) [2];
		}
		if ( ( position[i] ) [0]>xMax )
		{
			xMax= ( position[i] ) [0];
		}
		if ( ( position[i] ) [1]>yMax )
		{
			yMax= ( position[i] ) [1];
		}
		if ( ( position[i] ) [2]>zMax )
		{
			zMax= ( position[i] ) [2];
		}
	}
	ODEBUG2 ( ".......axisAlignedBoundingBox. " );

	return KD_OK;
}
// ==========================================================================
ktStatus ChppDynamicObstacleKCD::setAbsolutePosition ( const CkitMat4 &i_matrix )
{
	if ( attKCDObject.lock()->setAbsolutePosition ( i_matrix ) !=KD_OK )
	{
		ODEBUG1 ( "ERROR : ChppDynamicObstacleKCD::setAbsolutePosition. ");
		return KD_OK;
	}
	obstacleIsModified ();
	return KD_OK;
}
