/*
   Copyright (c) CNRS,
   @author : Alireza NAKHAEI   anakhaei@laas.fr

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
modification,
   are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
notice,
   this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
   * Neither the name of the <ORGANIZATION> nor the names of its
contributors
   may be used to endorse or promote products derived from this software
without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS
   OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY
   AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER
   OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
   IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/


#include <iostream>
#include "hppDynamicObstacle/hppPolyhedronMap.h"

// Select verbosity at configuration by setting CXXFLAGS="... -DDEBUG=[1 or 2]"
#if DEBUG==2
#define ODEBUG2(x) std::cout << "hppPolyhedronMap:" << x << std::endl
#define ODEBUG1(x) std::cerr << "hppPolyhedronMap:" << x << std::endl
#elif DEBUG==1
#define ODEBUG2(x)
#define ODEBUG1(x) std::cerr << "hppPolyhedronMap:" << x << std::endl
#else
#define ODEBUG2(x)
#define ODEBUG1(x)
#endif


unsigned int ChppPolyhedronMap::nbObject = 0;

ChppPolyhedronMap::ChppPolyhedronMap ( const std::string &i_name, double inSize )
{
	attResolution=inSize;
	attBoxVector.resize ( X_MAP*Y_MAP*Z_MAP );
	attPolydedronMap = CkppKCDAssembly::create ( i_name );
	attColor=CkppColor ( 0.1f, 0.9f, 0.1f, 0.4f );
	attAssemblyName= i_name;

	nbObject++;
	ODEBUG2 ( "Constructor of ChppPolyhedronMap. number of objects: "<< nbObject );
	cout << " Constructor of ChppPolyhedronMap: " << i_name <<endl;
}
//===========================================================================
ChppPolyhedronMap::~ChppPolyhedronMap()
{
	nbObject--;
	ODEBUG2 ( "---------------------------------------------------------------------------------------------------" );
	ODEBUG2 ( "      Destructor of class ChppPolyhedronMap. number of objects:" << nbObject );
	ODEBUG2 ( "--------------------------------------------------------------------------------------------------" );
}
// ==========================================================================
ChppPolyhedronMapShPtr ChppPolyhedronMap::create ( const std::string &i_name, double inSize )
{
	ChppPolyhedronMap*  flatPtr = new ChppPolyhedronMap ( i_name, inSize );
	ChppPolyhedronMapShPtr flatShPtr ( flatPtr );
	ChppPolyhedronMapWkPtr flatWkPtr ( flatShPtr );

	if ( flatPtr->init ( flatWkPtr ) != KD_OK ) flatShPtr.reset();

	return flatShPtr;
}
// ==========================================================================
ktStatus ChppPolyhedronMap::init ( const ChppPolyhedronMapWkPtr &inWeakPtr )
{
	attWeakPtr = inWeakPtr;

	return KD_OK;
}
// ==========================================================================
ktStatus ChppPolyhedronMap::addPolyhedron ( int i, int j, int k )
{
	std::vector<double> pos;
	pos.resize ( 3 );
	pos=ijkToPos ( i,j,k );
	if ( addPolyhedron ( pos.at ( 0 ),pos.at ( 1 ), pos.at ( 2 ) ) !=KD_OK )
	{
		ODEBUG1 ( "ERROR:  ChppPolyhedronMap::removePolyhedron(int) for " << i<< "," << j<< "," << k );
		return KD_ERROR;
	}
	return KD_OK;

}
// ==========================================================================
ktStatus ChppPolyhedronMap::addPolyhedron ( double x, double y, double z )
{
	//ODEBUG2 ( "addPolyhedron(double) in " << x << " , " <<y  <<" , " << z << " ...." );
	int N= posToID ( x, y, z );
	char* boxName = "Box";
	char buffName[50];
	sprintf ( buffName, "%s%d", boxName, N );
	std::string oss= std::string ( buffName );

	if ( ! attBoxVector.at ( N ) )
	{
		attBoxVector.at ( N ) = createBox ( oss , x, y, z );
	}

	attPolydedronMap->addChildComponent ( attBoxVector.at ( N ) );

	//ODEBUG2 ( " .......addPolyhedron(double) in was done" << x << " , " <<y  <<" , " << z );

	return KD_OK;
}
// ==========================================================================
ktStatus ChppPolyhedronMap::removePolyhedron ( int i, int j, int k )
{

	std::vector<double> pos;
	pos.resize ( 3 );
	pos=ijkToPos ( i,j,k );
	if ( removePolyhedron ( pos.at ( 0 ),pos.at ( 1 ), pos.at ( 2 ) ) !=KD_OK )
	{
		ODEBUG1 ( "ERROR: ChppPolyhedronMap::removePolyhedron(int) for " << i<< "," << j<< "," << k );
		return KD_ERROR;
	}
	return KD_OK;

}
// ==========================================================================
ktStatus ChppPolyhedronMap::removePolyhedron ( double x, double y, double z )
{
	//ODEBUG2 ( "removePolyhedron(double) in " << x << " , " <<y  <<" , " << z << " ...." );

	int N= posToID ( x, y, z );

	//unsigned int index;
	if ( attBoxVector.at ( N ) )
	{
		ODEBUG2 ( "DEBUG 1 : " << N );
		attPolydedronMap->removeChildComponent ( attBoxVector.at ( N ) );
		ODEBUG2 ( "DEBUG 2 : " << N );

		//attBoxVector.erase ( attBoxVector.begin() +N-1 );
	}
	//ODEBUG2 ( "-------------removePolyhedron(double) was done " << x << " , " <<y  <<" , " << z );

	return KD_OK;
}
// ==========================================================================

std::vector<double> ChppPolyhedronMap::ijkToPos ( int i, int j, int k )
{
	std::vector<double> pos;
	pos.resize ( 3 );
	pos.at ( 0 ) = ( i-X_MAP/2 ) *attResolution;
	pos.at ( 1 ) = ( i-Y_MAP/2 ) *attResolution;
	pos.at ( 2 ) = ( i-Z_MAP/2 ) *attResolution;


	return pos;
}
// ==========================================================================
int ChppPolyhedronMap::posToID ( double x, double y, double z )
{
	int i = ( int ) ceil ( x/ ( double ) attResolution ) +X_MAP/2;
	int j = ( int ) ceil ( y/ ( double ) attResolution ) +Y_MAP/2;
	int k = ( int ) ceil ( z/ ( double ) attResolution ) +Z_MAP/2;

	return ( ( k-1 ) *X_MAP*Y_MAP+ ( j-1 ) *X_MAP+ ( i-1 ) );
}
// ==========================================================================
std::vector<double> ChppPolyhedronMap::idToPos ( int id )
{
	std::vector<double> pos;
	pos.resize ( 3 );

	int NbOfVoxelInLine=X_MAP;
	int NbOfVoxelInLayer=X_MAP*Y_MAP;

	int j=0;
	int Nb, lineNb, layerNb;
	double xx,yy,zz;
	j++;

	Nb = id+1;
	lineNb=1;
	layerNb=1;
	if ( Nb> NbOfVoxelInLayer )
	{
		layerNb = ( Nb/NbOfVoxelInLayer ) +1;
		Nb=Nb%NbOfVoxelInLayer;
		if ( Nb==0 )
		{
			layerNb=layerNb-1;
			Nb=NbOfVoxelInLayer;
		}
	}

	if ( Nb> NbOfVoxelInLine )
	{
		lineNb = ( Nb/NbOfVoxelInLine ) +1;
		Nb=Nb%NbOfVoxelInLine;
		if ( Nb==0 )
		{
			lineNb=lineNb-1;
			Nb=NbOfVoxelInLine;
		}
	}

	xx= ( Nb-X_MAP/2 ) *attResolution-attResolution/2;
	yy= ( lineNb-Y_MAP/2 ) *attResolution-attResolution/2;
	zz= ( layerNb-Z_MAP/2 ) *attResolution-attResolution/2;

	pos [ 0 ] =xx;
	pos [ 1 ] =yy;
	pos [ 2 ] =zz;

	return pos;
}
// ==========================================================================
CkppKCDBoxShPtr ChppPolyhedronMap::createBox ( const std::string &inName, const double x, const double y, const double z )
{
	CkitMat4 absPos;
	CkppKCDBoxShPtr polyBox = CkppKCDBox::create ( inName, attResolution, attResolution,attResolution );
	absPos.translate ( x, y, z );
	polyBox->setAbsolutePosition ( absPos );
	polyBox->diffuseColor ( attColor );
	return polyBox;
}
// ==========================================================================
CkppKCDAssemblyShPtr ChppPolyhedronMap::getPolyhedronMap()
{
	return attPolydedronMap;
}
// ==========================================================================
double ChppPolyhedronMap::getSize()
{
	return attResolution;
}
// ==========================================================================
void ChppPolyhedronMap::diffuseColor ( CkppColor inColor )
{
	attColor=inColor;

}
// ==========================================================================
void ChppPolyhedronMap::clear( )
{
	ODEBUG2 ( "Clearing Map ..." );
	attPolydedronMap.reset();
	attPolydedronMap = CkppKCDAssembly::create ( attAssemblyName );
}
